/*
 * Copyright Mesa3D Contributors
 * SPDX-License-Identifier: MIT
 *
 * Autogenerated file, do not edit
 */

#include "compiler/nir/nir.h"
#include "compiler/nir/nir_builder.h"

#include "util/u_printf.h"

#ifdef __cplusplus
extern "C" {
#endif
nir_def *
blorp_check_in_bounds(nir_builder *b, nir_def *bounds_rect, nir_def *pos);

nir_def *
blorp_check_in_bounds(nir_builder *b, nir_def *bounds_rect, nir_def *pos)
{
   /*
    block b0:  // preds:
    32x4  %0 = @load_param (param_idx=0)
    32x2  %1 = @load_param (param_idx=1)
    1     %2 = uge %1.x, %0.x
    1     %3 = ult %1.x, %0.y
    1     %4 = uge %1.y, %0.z
    1     %5 = iand %4, %3
    1     %6 = iand %5, %2
    1     %7 = ult %1.y, %0.w
    1     %8 = iand %6, %7
               @bindgen_return (%8)
               // succs: b1
    block b1:
   */
   static const uint32_t impl_0_nir[43] = {
    0x0000000b, 0x00000028, 0x00000000, 0x00000000,
    0x00000000, 0x00000002, 0x00002004, 0x00000300,
    0x00000020, 0x00000000, 0x00002002, 0x00000280,
    0x00000020, 0x00000000, 0x00000001, 0x00000000,
    0x00000001, 0x00000000, 0x0000000a, 0xd0001ae4,
    0xc8011ae4, 0x242e0000, 0x00000000, 0x00020003,
    0x242e3200, 0x00000000, 0x00020003, 0x242e0480,
    0x00000000, 0x00020003, 0x246a0000, 0x00000000,
    0x00050006, 0x00000000, 0x00040007, 0x242e3680,
    0x00000000, 0x00020003, 0x242a0000, 0x00000000,
    0x00090008, 0x00000324, 0x00000028,
};
   nir_def *args[2] = { bounds_rect, pos };
   return nir_call_serialized(b, impl_0_nir, sizeof(impl_0_nir), args);
}

#ifdef __cplusplus
} /* extern C */
#endif
namespace {
   struct vtn_bindgen_dummy {
      vtn_bindgen_dummy() {
         /* Format strings:
          *
          */
const uint32_t printf_0_blob[1] = {
    0x00000000,
};
         u_printf_singleton_init_or_ref();
         u_printf_singleton_add_serialized((const void*)printf_0_blob, sizeof(printf_0_blob));
      }

      ~vtn_bindgen_dummy() {
         u_printf_singleton_decref();
      }
   };

   static vtn_bindgen_dummy vtn_bindgen_dummy_instance;
}
