OPT_BOOL(nohiz, false, "Disable hierarchical zbuffer")
OPT_BOOL(nozmask, false, "Disable zbuffer compression")
OPT_BOOL(ieeemath, false, "Force IEEE math rules and opcodes where applicable")
OPT_BOOL(ffmath, false, "Force FF math rules and opcodes where applicable")

#undef OPT_BOOL
