/*
 * Copyright © 2013 Rob Clark <robclark@freedesktop.org>
 * SPDX-License-Identifier: MIT
 *
 * Authors:
 *    Rob Clark <robclark@freedesktop.org>
 */

#ifndef FD2_SCREEN_H_
#define FD2_SCREEN_H_

#include "pipe/p_screen.h"

void fd2_screen_init(struct pipe_screen *pscreen);

#endif /* FD2_SCREEN_H_ */
